/* Copyright 2023 Kevin Ryde

   This file is part of pari-primecount.

   pari-primecount is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   pari-primecount is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with pari-primecount.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Set WANT_ASSERT to 1 to enable self-checks.
   Set WANT_DEBUG to 1 for some rough development prints.
*/
#define WANT_ASSERT  0
#define WANT_DEBUG   0

#if ! WANT_ASSERT
#define NDEBUG
#endif

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <inttypes.h>
#include <pari/pari.h>
#include <primecount.h>


/* ---------------------------------------------------------------------- */
/* Generic */

#if WANT_DEBUG
#define DEBUG(expr)  do { expr; } while (0)
#else
#define DEBUG(expr)  do {       } while (0)
#endif


/* ---------------------------------------------------------------------- */
/* PARI Generic */

/* capped at INT_MIN, INT_MAX */
static int
gen_to_int_capped (GEN n)
{
  if (gcmpgs(n,INT_MAX) > 0) return INT_MAX;
  if (gcmpgs(n,INT_MIN) < 0) return INT_MIN;
  return(gtos(n));
}

static GEN
int64_to_gen (int64_t x)
{
  DEBUG(printf("int64_to_gen() %"PRId64"\n", x));
#if BITS_IN_LONG >= 64
  return (stoi(x));

#elif BITS_IN_LONG == 32
  /* some primecount functions return -1 so allow negative x */
  int neg = (x < 0);
  uint64_t u = (neg ? -x : x);
  uint64_t low_mask = (((uint64_t) 1) << 32) - 1;
  GEN g = uu32toi(u >> 32, u & low_mask);
  return (neg ? gneg(g) : g);

#else
#error Unrecognised BITS_IN_LONG
#endif
}

/* GEN x is t_INT in the range 0 .. INT64_MAX.
   Return it as an int64_t. */
static int64_t
gen_to_int64_pos_nocheck (GEN x)
{
  DEBUG(pari_printf("gen_to_int64_pos_nocheck() %Ps type=%ld\n", x, typ(x));
        if(typ(x) == t_INT) pari_printf("  hex %Px\n", x));
  assert (typ(x) == t_INT);
  assert (gsigne(x) >= 0);

#if BITS_IN_LONG >= 64
  return (gtos(x));

#elif BITS_IN_LONG == 32
  long xlen = lgefint(x);
  long *xptr = int_LSW(x);
  assert (xlen <= 4);
  unsigned long lo = (xlen >= 3 ? *xptr            : 0);
  unsigned long hi = (xlen >= 4 ? *int_nextW(xptr) : 0);
  DEBUG(printf("  xlen %ld words lo=%#lx hi=%#lx\n", xlen, lo, hi));
  return ((int64_t) hi << 32) | lo;

#else
#error Unrecognised BITS_IN_LONG
#endif
}

/* ---------------------------------------------------------------------- */
/* Inits */

static GEN two_pow_63;       /* 2^63 */
static GEN pi_max_x;         /* max for primecount_pi() */
static GEN nth_prime_max_n;  /* max for primecount_nth_prime() */

void
pari_primecount_init(void)
{
  DEBUG(setbuf(stdout,NULL);
        printf("pari_primecount_init()\n");
        printf("  BITS_IN_LONG    = %d\n", BITS_IN_LONG));

  static char init_done = 0;
  if (! init_done) {
    init_done = 1;

    /* nth_prime_max_n per print in primecount error message,
       and which is prime(216289611853439384) biggest fits int64_t
       and per OEIS A014234 Largest prime <= 2^n. */
    nth_prime_max_n = gclone(gp_read_str("216289611853439384"));
    two_pow_63 = gclone(shifti(gen_1,63));
    pi_max_x = gclone(strtoi(primecount_get_max_x()));

    DEBUG(pari_printf("  nth_prime_max_n = %Ps\n", nth_prime_max_n);
          pari_printf("  two_pow_63      = %Ps\n", two_pow_63);
          pari_printf("  pi_max_x        = %Ps\n", pi_max_x));
  }
}


/* ---------------------------------------------------------------------- */

GEN
pari_primecount_version (void)
{
  return(strtoGENstr(primecount_version()));
}

/* primecount_get_num_threads() and primecount_set_num_threads()
   act on "int" rather than long, so a direct install() might
   be wrong if int!=long.
*/
GEN
pari_primecount_get_num_threads (void)
{
  return(stoi(primecount_get_num_threads()));
}
void
pari_primecount_set_num_threads (GEN n)
{
  /* primecount takes n<1 to mean n=1
     and n > actual num cores to mean n = actual num cores.
     In that spirit, cap GEN -> int rather than e_OVERFLOW
  */
  primecount_set_num_threads(gen_to_int_capped(n));
}


/* ---------------------------------------------------------------------- */

/* const char* primecount_get_max_x();
   Maximum arg for primecount_pi(). */
GEN
pari_primecount_get_max_x (void)
{
  return(pi_max_x);
}

/* int64_t primecount_pi(int64_t x);
   int primecount_pi_str(const char* x, char* res, size_t len);
*/
GEN
pari_primecount_pi (GEN x)
{
  /* primecount_pi() and primecount_pi_str() segfault on x < 0.
     Here return 0 like PARI primepi().

     primecount_pi_str() prints a message and segfaults on x too big.
     Throw PARI e_OVERFLOW for that.
  */

  x = gfloor(x);
  if (typ(x) != t_INT) pari_err_TYPE("pari_primecount_pi",x);
  if (gsigne(x) <= 0)
    return (gen_0);
  if (gcmp(x, pi_max_x) > 0)
    pari_err_OVERFLOW("bigger than max_x");

  if (gcmp(x, two_pow_63) < 0) {
    DEBUG(printf("pari_primecount_pi() by int64_t\n"));
    return (int64_to_gen(primecount_pi(gen_to_int64_pos_nocheck(x))));

  } else {
    char *str = itostr(x);
    DEBUG(printf("pari_primecount_pi() by str %s\n", str));
    char res[32];
    int ret = primecount_pi_str(str, res, sizeof(res));
    if (ret == -1)
      pari_err_OVERFLOW("primecount_pi_str");
    return(gp_read_str(res));
  }
}

/* int64_t primecount_phi(int64_t x, int64_t a); */
GEN
pari_primecount_phi (GEN x, GEN a)
{
  /* pari_primecount_phi() allows negative "x" and treats it as
     no numbers 1..x.  Allow bignum negative x in that spirit.

     pari_primecount_phi() allows negative "a" and treats it as
     no primes.  Allow bignum negative "a" in that spirit.

     Take gfloor(x) in the same style as pari_primecount_pi() since
     numbers <= x.  But "a" is an prime index so must be integer.
  */

  x = gfloor(x);
  if (typ(x) != t_INT) pari_err_TYPE("pari_primecount_pi [x]",x);
  if (typ(a) != t_INT) pari_err_TYPE("pari_primecount_pi [a]",a);
  if (gsigne(x) <= 0 || gsigne(a) <= 0)
    return (gen_0);
  if (gcmp(x,two_pow_63) >= 0 || gcmp(a,two_pow_63) >= 0)
    pari_err_OVERFLOW("primecount_phi");

  return(int64_to_gen(primecount_phi(gen_to_int64_pos_nocheck(x),
                                     gen_to_int64_pos_nocheck(a))));
}

/* int64_t primecount_nth_prime(int64_t n); */
GEN
pari_primecount_nth_prime (GEN n)
{
  /* primecount_nth_prime() on n<1 prints an error message.
     Throw PARI e_DOMAIN for that.

     primecount_nth_prime() on n > nth_prime_max_n prints an error
     message since resulting prime too big for int64_t.
     Throw PARI e_OVERFLOW for that. */

  DEBUG(pari_printf("pari_primecount_nth_prime() n=%Ps\n", n));

  if (typ(n) != t_INT) pari_err_TYPE("primecount_nth_prime",n);
  if (gsigne(n) <= 0)
    pari_err_DOMAIN("primecount_nth_prime", "n", "<",gen_1,n);
  if (gcmp(n,nth_prime_max_n) > 0)
    pari_err_OVERFLOW("primecount_nth_prime");

  int64_t prime = primecount_nth_prime(gen_to_int64_pos_nocheck(n));
  return(int64_to_gen(prime));
}
